import logging
from copy import deepcopy
from functools import wraps

import numpy

__all__ = [
    "dtypes",
    "block",
    "check_parameters",
]


dtypes = {
    "PARAMETERS": {
        "title": "str",
        "eos": "str",
        "n_component": "int",
        "n_phase": "int",
        "n_phase": "int",
        "flac": "bool",
        "creep": "bool",
        "porosity_model": "int",
        "isothermal": "bool",
        "start": "bool",
        "nover": "bool",
        "endfi": "bool",
        "rocks": "dict",
        "rocks_order": "array_like",
        "options": "dict",
        "extra_options": "dict",
        "more_options": "dict",
        "selections": "dict",
        "extra_selections": "array_like",
        "solver": "dict",
        "generators": "dict",
        "times": "scalar_array_like",
        "element_history": "array_like",
        "connection_history": "array_like",
        "generator_history": "array_like",
        "default": "dict",
    },
    "ROCKS": {
        "density": "scalar",
        "porosity": "scalar",
        "permeability": "scalar_array_like",
        "conductivity": "scalar",
        "specific_heat": "scalar",
        "compressibility": "scalar",
        "expansion": "scalar",
        "conductivity_dry": "scalar",
        "tortuosity": "scalar",
        "b_coeff": "scalar",
        "xkd3": "scalar",
        "xkd4": "scalar",
        "relative_permeability": "dict",
        "capillarity": "dict",
        "permeability_model": "dict",
        "equivalent_pore_pressure": "dict",
    },
    "MODEL": {"id": "int", "parameters": "array_like",},
    "PARAM": {
        "n_iteration": "int",
        "n_cycle": "int",
        "n_second": "int",
        "n_cycle_print": "int",
        "verbosity": "int",
        "temperature_dependence_gas": "scalar",
        "effective_strength_vapor": "scalar",
        "t_ini": "scalar",
        "t_max": "scalar",
        "t_steps": "scalar_array_like",
        "t_step_max": "scalar",
        "t_reduce_factor": "scalar",
        "gravity": "scalar",
        "mesh_scale_factor": "scalar",
        "eps1": "scalar",
        "eps2": "scalar",
        "w_upstream": "scalar",
        "w_newton": "scalar",
        "derivative_factor": "scalar",
        "incon": "array_like",
    },
    "MOP": {i + 1: "int" for i in range(24)},
    "MOMOP": {i + 1: "int" for i in range(40)},
    "SELEC": {i + 1: "int" for i in range(16)},
    "SOLVR": {
        "id": "int",
        "z_precond": "str",
        "o_precond": "str",
        "rel_iter_max": "scalar",
        "eps": "scalar",
    },
    "GENER": {
        "type": "str",
        "times": "scalar_array_like",
        "rates": "scalar_array_like",
        "specific_enthalpy": "scalar_array_like",
        "layer_thickness": "scalar_array_like",
    },
}


str_to_dtype = {
    "int": (int, numpy.int32),
    "float": (float, numpy.float32, numpy.float64),
    "str": (str,),
    "bool": (bool,),
    "array_like": (list, tuple, numpy.ndarray),
    "dict": (dict,),
    "scalar": (int, float, numpy.int32, numpy.float32, numpy.float64),
    "scalar_array_like": (
        int,
        float,
        list,
        tuple,
        numpy.int32,
        numpy.float32,
        numpy.float64,
        numpy.ndarray,
    ),
}


def block(keyword, multi=False, noend=False):
    """
    Decorator for block writing functions.
    """

    def decorator(func):
        from .._common import header

        @wraps(func)
        def wrapper(*args, **kwargs):
            head_fmt = "{:5}{}" if noend else "{:5}{}\n"
            out = [head_fmt.format(keyword, header)]
            out += func(*args, **kwargs)
            out += ["\n"] if multi else []

            return out

        return wrapper

    return decorator


def check_parameters(input_types, keys=None, is_list=False):
    """
    Decorator to check input parameters.
    """

    def _check_parameters(params, keys=None):
        for k, v in params.items():
            # Check whether parameters contain unknown keys
            # Log error if it does and skip
            if k not in input_types.keys():
                logging.warning(
                    "Unknown key '{}'{}. Skipping.".format(
                        k, " in {}".format(keys) if keys else ""
                    )
                )
                continue

            # Check input types
            input_type = str_to_dtype[input_types[k]]
            assert v is None or isinstance(
                v, input_type
            ), "Invalid type for parameter '{}' {}(expected {}).".format(
                k, "in {} ".format(keys) if keys else "", input_types[k],
            )

    keys = [keys] if isinstance(keys, str) else keys

    def decorator(func):
        @wraps(func)
        def wrapper(parameters):
            if not keys:
                _check_parameters(parameters)
            else:
                params = deepcopy(parameters[keys[0]])
                keys_str = "['{}']".format(keys[0])
                if is_list:
                    for k, v in params.items():
                        tmp = keys_str
                        tmp += "['{}']".format(k)
                        for key in keys[1:]:
                            v = v[key]
                            tmp += "['{}']".format(key)
                        _check_parameters(v, tmp)
                else:
                    for key in keys[1:]:
                        params = params[key]
                        keys_str += "['{}']".format(key)
                    _check_parameters(params, keys_str)

            out = func(parameters)

            return out

        return wrapper

    return decorator



import numpy

from ..mesh._mesh import Cells, Mesh

__all__ = [
    "triangulate",
]


def triangulate(points, material="dfalt"):
    """
    Generate a triangulated mesh from cloud points.

    Parameters
    ----------
    points : array_like (n_points, 3)
        Cooordinates of points.
    material : str, optional, default 'dfalt'
        Default material name.

    Returns
    -------
    toughio.Mesh
        Output triangulated mesh.
    """
    try:
        from scipy.spatial import Delaunay
    except ImportError:
        raise ModuleNotFoundError(
            "Triangulation requires scipy >= 0.9 to be installed."
        )

    assert isinstance(points, (list, tuple, numpy.ndarray))
    assert numpy.ndim(points) == 2
    assert numpy.shape(points)[1] in {2, 3}
    assert isinstance(material, str)

    points = numpy.asarray(points)
    connectivity = Delaunay(points).simplices
    cell_type = "triangle" if points.shape[1] == 2 else "tetra"
    cells = [Cells(cell_type, connectivity)]
    mesh = Mesh(points, cells)
    mesh.add_cell_data("material", numpy.ones(mesh.n_cells, dtype=int))
    mesh.field_data[material] = numpy.array([1, 3])

    return mesh



from __future__ import with_statement

import json

__all__ = [
    "read",
    "write",
]


def read(filename):
    """
    Import json TOUGH input file.
    
    Parameters
    ----------
    filename : str
        Input file name.
    
    Returns
    -------
    dict
        TOUGH input parameters.
    """

    def to_int(data):
        """
        Return dict with integer keys instead of strings.
        """
        return {int(k): data[k] for k in sorted(data.keys())}

    with open(filename, "r") as f:
        parameters = json.load(f)

    keys = {"extra_options", "more_options", "selections"}
    for key in keys:
        if key in parameters.keys():
            parameters[key] = to_int(parameters[key])

    return parameters


def write(filename, parameters):
    """
    Export TOUGH parameters to json.

    Parameters
    ----------
    filename : str
        Output file name.
    parameters : dict
        Parameters to export.
    """
    with open(filename, "w") as f:
        json.dump(parameters, f, indent=4)



from __future__ import with_statement

import collections

import numpy

from . import json, tough

__all__ = [
    "read_input",
    "write_input",
    "read_output",
    "read_save",
]


_extension_to_filetype = {
    ".json": "json",
}


def _filetype_from_filename(filename):
    """
    Determine file type from its extension.
    """
    import os

    ext = os.path.splitext(filename)[1].lower()
    return (
        _extension_to_filetype[ext] if ext in _extension_to_filetype.keys() else "tough"
    )


def read_input(filename, file_format="json", **kwargs):
    """
    Read TOUGH input file.

    Parameters
    ----------
    filename : str
        Input file name.
    file_format : str ('tough', 'json'), optional, default 'json'
        Input file format.
    
    Returns
    -------
    dict
        TOUGH input parameters.
    """
    assert isinstance(filename, str)
    assert file_format in {"tough", "json"}

    fmt = file_format if file_format else _filetype_from_filename(filename)

    format_to_reader = {
        "tough": (tough, (), {}),
        "json": (json, (), {}),
    }
    interface, args, default_kwargs = format_to_reader[fmt]
    _kwargs = default_kwargs.copy()
    _kwargs.update(kwargs)
    return interface.read(filename, *args, **_kwargs)


def write_input(filename, parameters, file_format="tough", **kwargs):
    """
    Write TOUGH input file.

    Parameters
    ----------
    filename : str
        Output file name.
    parameters : dict
        Parameters to export.
    file_format : str ('tough', 'json'), optional, default 'tough'
        Output file format.
    """
    assert isinstance(filename, str)
    assert isinstance(parameters, dict)
    assert file_format in {"tough", "json"}

    fmt = file_format if file_format else _filetype_from_filename(filename)

    format_to_writer = {
        "tough": (tough, (), {}),
        "json": (json, (), {}),
    }
    interface, args, default_kwargs = format_to_writer[fmt]
    _kwargs = default_kwargs.copy()
    _kwargs.update(kwargs)
    interface.write(filename, parameters, *args, **_kwargs)


def read_output(filename):
    """
    Read TOUGH output file for each time step.

    Parameters
    ----------
    filename : str
        Input file name.

    Returns
    -------
    list of namedtuple
        List of namedtuple (time, labels, data) for each time step.
    """
    assert isinstance(filename, str)

    with open(filename, "r") as f:
        # Read header
        line = f.readline().replace('"', "")
        headers = [l.strip() for l in line.split(",")]

        # Skip second line (unit)
        line = f.readline()

        # Read data
        times, labels, variables = [], [], []
        line = f.readline().replace('"', "").strip()
        while line:
            line = line.split(",")

            # Time step
            if line[0].startswith("TIME"):
                line = line[0].split()
                times.append(float(line[-1]))
                variables.append([])
                labels.append([])

            # Output
            else:
                labels[-1].append(line[0].strip())
                variables[-1].append([float(l.strip()) for l in line[1:]])

            line = f.readline().strip().replace('"', "")

    Output = collections.namedtuple("Output", ["time", "labels", "data"])

    return [
        Output(
            time, label, {k: v for k, v in zip(headers[1:], numpy.transpose(variable))}
        )
        for time, label, variable in zip(times, labels, variables)
    ]


def read_save(filename):
    """
    Read TOUGH SAVE file.

    Parameters
    ----------
    filename : str
        Input file name.

    Returns
    -------
    list of namedtuple
        SAVE data as namedtuple (labels, data).

    Note
    ----
    Does not support porosity, USERX and hysteresis values yet.
    """

    def str2float(s):
        """
        Convert primary variables string to float.
        """
        s = s.lower()
        significand, exponent = s[:-4], s[-4:].replace("e", "")
        return float("{}e{}".format(significand, exponent))

    assert isinstance(filename, str)

    with open(filename, "r") as f:
        # Check first line
        line = f.readline()
        assert line.startswith("INCON"), "Invalid SAVE file '{}'.".format(filename)

        # Read data
        labels, variables = [], []
        line = f.readline().strip()
        while line:
            # Label
            line = line.split()
            labels.append(line[0])

            # Primary variables
            line = f.readline().strip().split()
            variables.append([str2float(l) for l in line])

            line = f.readline().strip()
            if line.startswith("+++"):
                break

    Save = collections.namedtuple("Save", ["labels", "data"])

    return Save(
        labels,
        {"X{}".format(i + 1): x for i, x in enumerate(numpy.transpose(variables))},
    )



from __future__ import division, unicode_literals, with_statement

import time

import numpy

from ...__about__ import __version__ as version
from .._common import meshio_data
from .._mesh import Cells, Mesh

__all__ = [
    "read",
    "write",
]


meshio_only = {
    "tetra": "tetra",
    "tetra10": "tetra",
    "pyramid": "pyramid",
    "pyramid13": "pyramid",
    "wedge": "wedge",
    "wedge12": "wedge",
    "wedge15": "wedge",
    "wedge18": "wedge",
    "hexahedron": "hexahedron",
    "hexahedron20": "hexahedron",
    "hexahedron24": "hexahedron",
    "hexahedron27": "hexahedron",
}


numnodes_to_meshio_type = {
    4: "tetra",
    5: "pyramid",
    6: "wedge",
    8: "hexahedron",
}
meshio_type_to_numnodes = {v: k for k, v in numnodes_to_meshio_type.items()}


meshio_to_flac3d_type = {
    "tetra": "T4",
    "pyramid": "P5",
    "wedge": "W6",
    "hexahedron": "B8",
}


flac3d_to_meshio_order = {
    "tetra": [0, 1, 2, 3],
    "pyramid": [0, 1, 4, 2, 3],
    "wedge": [0, 1, 3, 2, 4, 5],
    "hexahedron": [0, 1, 4, 2, 3, 6, 7, 5],
}


meshio_to_flac3d_order = {
    "tetra": [0, 1, 2, 3],
    "pyramid": [0, 1, 3, 4, 2],
    "wedge": [0, 1, 3, 2, 4, 5],
    "hexahedron": [0, 1, 3, 4, 2, 7, 5, 6],
}


meshio_to_flac3d_order_2 = {
    "tetra": [0, 2, 1, 3],
    "pyramid": [0, 3, 1, 4, 2],
    "wedge": [0, 2, 3, 1, 5, 4],
    "hexahedron": [0, 3, 1, 4, 2, 5, 7, 6],
}


def read(filename):
    """
    Read FLAC3D f3grid grid file (only ASCII).
    """
    with open(filename, "r") as f:
        out = read_buffer(f)
    return out


def read_buffer(f):
    """
    Read ASCII file line by line. Use combination of readline, tell and
    seek since we need to rewind to the previous line when we read last
    data line of a ZGROUP section.
    """
    points = []
    point_ids = {}
    cells = []
    mapper = {}
    field_data = {}
    slots = set()

    pidx = 0
    zidx = 0
    count = 0
    line = f.readline().rstrip().split()
    while line:
        if line[0] == "G":
            pid, point = _read_point(line)
            points.append(point)
            point_ids[pid] = pidx
            pidx += 1
        elif line[0] == "Z":
            cid, cell = _read_cell(line, point_ids)
            cell_type = numnodes_to_meshio_type[len(cell)]
            if len(cells) > 0 and cell_type == cells[-1][0]:
                cells[-1][1].append(cell)
            else:
                cells.append((cell_type, [cell]))
            mapper[cid] = [count]
            count += 1
        elif line[0] == "ZGROUP":
            name, data, slot = _read_zgroup(f, line)
            zidx += 1
            for cid in data:
                mapper[cid].append(zidx)
            field_data[name] = numpy.array([zidx, 3])
            slots.add(slot)
            if len(slots) > 1:
                raise AssertionError("Multiple slots are not supported.")
        line = f.readline().rstrip().split()

    if zidx:
        n_cells = numpy.cumsum([len(c[1]) for c in cells])
        cell_data = numpy.empty(n_cells[-1], dtype=int)
        for cid, zid in mapper.values():
            cell_data[cid] = zid
        cell_data = {"flac3d:zone": numpy.split(cell_data, n_cells[:-1])}
    else:
        cell_data = {}

    return Mesh(
        points=numpy.array(points),
        cells=[(k, numpy.array(v)[:, flac3d_to_meshio_order[k]]) for k, v in cells],
        cell_data=cell_data,
        field_data=field_data,
    )


def _read_point(line):
    """
    Read point coordinates.
    """
    return int(line[1]), [float(l) for l in line[2:]]


def _read_cell(line, point_ids):
    """
    Read cell corners.
    """
    cell = [point_ids[int(l)] for l in line[3:]]
    if line[1] == "B7":
        cell.append(cell[-1])
    return int(line[2]), cell


def _read_zgroup(f, line):
    """
    Read cell group.
    """
    name = line[1].replace('"', "")
    data = []
    slot = "" if "SLOT" not in line else line[-1]

    i = f.tell()
    line = f.readline()
    while True:
        line = line.rstrip().split()
        if line and (line[0] not in {"*", "ZGROUP"}):
            data += [int(l) for l in line]
        else:
            f.seek(i)
            break
        i = f.tell()
        line = f.readline()
    return name, data, slot


def write(filename, mesh):
    """
    Write FLAC3D f3grid grid file (only ASCII).
    """
    if not any(c.type in meshio_only.keys() for c in mesh.cells):
        raise AssertionError("FLAC3D format only supports 3D cells.")

    with open(filename, "w") as f:
        f.write("* FLAC3D grid produced by toughio v{}\n".format(version))
        f.write("* {}\n".format(time.ctime()))
        f.write("* GRIDPOINTS\n")
        _write_points(f, mesh.points)
        f.write("* ZONES\n")
        _write_cells(f, mesh.points, mesh.cells)

        if mesh.cell_data:
            if set(list(mesh.cell_data)).intersection(meshio_data):
                f.write("* ZONE GROUPS\n")
                _write_cell_data(f, mesh.cells, mesh.cell_data, mesh.field_data)


def _write_points(f, points):
    """
    Write points coordinates.
    """
    for i, point in enumerate(points):
        f.write("G\t{:8}\t{:.14e}\t{:.14e}\t{:.14e}\n".format(i + 1, *point))


def _write_cells(f, points, cells):
    """
    Write zones.
    """
    zones = _translate_zones(points, cells)
    i = 1
    for meshio_type, zone in zones:
        fmt = "Z {} {} " + " ".join(["{}"] * zone.shape[1]) + "\n"
        for entry in zone + 1:
            f.write(fmt.format(meshio_to_flac3d_type[meshio_type], i, *entry))
            i += 1


def _translate_zones(points, cells):
    """
    Reorder meshio cells to FLAC3D zones. Four first points must form a right-handed
    coordinate system (outward normal vectors). Reorder corner points according to sign
    of scalar triple products.
    """
    # See <https://stackoverflow.com/a/42386330/353337>
    def slicing_summing(a, b, c):
        c0 = b[:, 1] * c[:, 2] - b[:, 2] * c[:, 1]
        c1 = b[:, 2] * c[:, 0] - b[:, 0] * c[:, 2]
        c2 = b[:, 0] * c[:, 1] - b[:, 1] * c[:, 0]
        return a[:, 0] * c0 + a[:, 1] * c1 + a[:, 2] * c2

    zones = []
    for key, idx in cells:
        if key not in meshio_only.keys():
            continue

        # Compute scalar triple products
        key = meshio_only[key]
        tmp = points[idx[:, meshio_to_flac3d_order[key][:4]].T]
        det = slicing_summing(tmp[1] - tmp[0], tmp[2] - tmp[0], tmp[3] - tmp[0])
        # Reorder corner points
        data = numpy.where(
            (det > 0)[:, None],
            idx[:, meshio_to_flac3d_order[key]],
            idx[:, meshio_to_flac3d_order_2[key]],
        )
        zones.append((key, data))

    return zones


def _write_cell_data(f, cells, cell_data, field_data):
    """
    Write zone groups.
    """
    zgroups, labels = _translate_zgroups(cells, cell_data, field_data)
    for k in sorted(zgroups.keys()):
        f.write('ZGROUP "{}"\n'.format(labels[k]))
        _write_zgroup(f, zgroups[k])


def _translate_zgroups(cells, cell_data, field_data):
    """
    Convert meshio cell_data to FLAC3D zone groups.
    """
    mat_data = None
    for k in cell_data.keys():
        if k in meshio_data:
            mat_data = k
            break

    num_cells = sum(len(c.data) for c in cells)
    zone_data = (
        numpy.concatenate(cell_data[mat_data])
        if mat_data
        else numpy.zeros(num_cells, dtype=int)
    )

    zgroups = {k: numpy.nonzero(zone_data == k)[0] + 1 for k in numpy.unique(zone_data)}

    labels = {k: str(k) for k in zgroups.keys()}
    labels[0] = "None"
    if field_data:
        labels.update({v[0]: k for k, v in field_data.items() if v[1] == 3})
    return zgroups, labels


def _write_zgroup(f, data, ncol=20):
    """
    Write zone group data.
    """
    nrow = len(data) // ncol
    lines = numpy.split(data, numpy.full(nrow, ncol).cumsum())
    for line in lines:
        if len(line):
            f.write(" {}\n".format(" ".join([str(l) for l in line])))



from . import mesh, meshmaker
from .__about__ import (
    __author__,
    __author_email__,
    __license__,
    __version__,
    __website__,
)
from ._io import read_input, read_output, read_save, write_input
from .mesh import Mesh
from .mesh import read as read_mesh
from .mesh import write as write_mesh

__all__ = [
    "Mesh",
    "mesh",
    "meshmaker",
    "read_input",
    "write_input",
    "read_output",
    "read_save",
    "read_mesh",
    "write_mesh",
    "__version__",
    "__author__",
    "__author_email__",
    "__website__",
    "__license__",
]



import collections
import logging
from copy import deepcopy

import meshio
import numpy

from ._common import get_local_index, get_meshio_version, meshio_data

__all__ = [
    "Cells",
    "Mesh",
    "from_meshio",
]


Cells = collections.namedtuple("Cells", ["type", "data"])


class Mesh(object):
    """
    toughio mesh.

    This class is updated following the latest :module:`meshio` version and
    brings backward compatibility with its previous versions.

    Parameters
    ----------
    points : array_like (n_points, 3)
        Cooordinates of points.
    cells : list of tuples (cell_type, data)
        Connectivity of cells.
    point_data : dict or None, optional, default None
        Point data arrays.
    cell_data : dict or None, optional, default None
        Cell data arrays.
    field_data : dict or None, optional, default None
        Field data names.
    point_sets : dict or None, optional, default None
        Sets of points.
    cell_sets : dict or None, optional, default None
        Sets of cells.
    """

    def __init__(
        self,
        points,
        cells,
        point_data=None,
        cell_data=None,
        field_data=None,
        point_sets=None,
        cell_sets=None,
    ):
        self.points = points
        self.cells = cells
        self.point_data = point_data if point_data else {}
        self.cell_data = cell_data if cell_data else {}
        self.field_data = field_data if field_data else {}
        self.point_sets = point_sets if point_sets else {}
        self.cell_sets = cell_sets if cell_sets else {}

    def __repr__(self):
        lines = [
            "<toughio mesh object>",
            "  Number of points: {}".format(len(self.points)),
        ]
        if len(self.cells) > 0:
            lines.append("  Number of cells:")
            for tpe, elems in self.cells:
                lines.append("    {}: {}".format(tpe, len(elems)))
        else:
            lines.append("  No cells.")

        if self.point_sets:
            lines.append("  Point sets: {}".format(", ".join(self.point_sets.keys())))

        if self.point_data:
            lines.append("  Point data: {}".format(", ".join(self.point_data.keys())))

        if self.cell_data:
            lines.append("  Cell data: {}".format(", ".join(self.cell_data.keys())))

        return "\n".join(lines)

    def extrude_to_3d(self, height=1.0, axis=2, inplace=True):
        """
        Convert a 2D mesh to 3D by extruding cells along given axis.

        Parameters
        ----------
        height : scalar or array_like, optional, default 1.0
            Height of extrusion.
        axis : int (0, 1 or 2), optional, default 2
            Axis along which extrusion is performed.
        inplace : bool, optional, default True
            If `False`, return a new :class:`toughio.Mesh`.

        Returns
        -------
        toughio.Mesh
            Extruded mesh (only if ``inplace == False``).
        """
        assert axis in [0, 1, 2], "axis must be 0, 1 or 2."
        mesh = self if inplace else deepcopy(self)
        height = [height] if isinstance(height, (int, float)) else height

        npts, nh = len(mesh.points), len(height)
        if mesh.points.shape[1] == 3:
            assert (
                len(set(mesh.points[:, axis])) == 1
            ), "Cannot extrude mesh along axis {}.".format(axis)
        else:
            mesh.points = numpy.column_stack((mesh.points, numpy.zeros(npts)))
            if axis != 2:
                mesh.points[:, [axis, 2]] = mesh.points[:, [2, axis]]

        extra_points = numpy.array(mesh.points)
        for h in height:
            extra_points[:, axis] += h
            mesh.points = numpy.vstack((mesh.points, extra_points))

        extruded_types = {
            "triangle": "wedge",
            "quad": "hexahedron",
        }
        cells = []
        for ic, c in enumerate(mesh.cells):
            if c.type in extruded_types.keys():
                extruded_type = extruded_types[c.type]
                nr, nc = c.data.shape
                cell = Cells(extruded_type, numpy.tile(c.data, (nh, 2)))
                for i in range(nh):
                    ibeg, iend = i * nr, (i + 1) * nr
                    cell.data[ibeg:iend, :nc] += i * npts
                    cell.data[ibeg:iend, nc:] += (i + 1) * npts
                cells.append(cell)

                if mesh.cell_data:
                    for k, v in mesh.cell_data.items():
                        v[ic] = numpy.tile(v[ic], nh)
        mesh.cells = cells

        if mesh.field_data:
            for k in mesh.field_data.keys():
                mesh.field_data[k][1] = 3

        if not inplace:
            return mesh

    def prune_duplicates(self, inplace=True):
        """
        Delete duplicate points and cells.

        Parameters
        ----------
        inplace : bool, optional, default True
            If `False`, return a new :class:`toughio.Mesh`.

        Returns
        -------
        toughio.Mesh
            Pruned mesh (only if ``inplace == False``).
        
        Note
        ----
        Does not preserve points order from original array in mesh.
        """
        mesh = self if inplace else deepcopy(self)
        cells = mesh.cells_dict

        # Prune duplicate points
        unique_points, pind, pinv = numpy.unique(
            mesh.points, axis=0, return_index=True, return_inverse=True,
        )
        if len(unique_points) < len(mesh.points):
            mesh.points = unique_points
            for k, v in mesh.point_data.items():
                mesh.point_data[k] = v[pind]
            for k, v in cells.items():
                cells[k] = pinv[v]

        # Prune duplicate cells
        for ic, (k, v) in enumerate(cells.items()):
            vsort = numpy.sort(v, axis=1)
            _, order = numpy.unique(vsort, axis=0, return_index=True)
            cells[k] = cells[k][order]
            if mesh.cell_data:
                for kk, vv in mesh.cell_data.items():
                    mesh.cell_data[kk][ic] = vv[ic][order]
        mesh.cells = cells

        if not inplace:
            return mesh

    def split(self, arr):
        """
        Split input array into subarrays for each cell block in mesh.

        Parameters
        ----------
        arr : array_like
            Input array.

        Returns
        -------
        list of array_like
            List of subarrays.
        """
        assert len(arr) == self.n_cells
        sizes = numpy.cumsum([len(c.data) for c in self.cells])

        return numpy.split(numpy.asarray(arr), sizes[:-1])

    def to_meshio(self):
        """
        Convert mesh to :class:`meshio.Mesh`.

        Returns
        -------
        meshio.Mesh
            Output mesh.
        """
        keys = ["points", "point_data", "field_data"]
        kwargs = {key: getattr(self, key) for key in keys}

        version = get_meshio_version()
        if version[0] >= 4:
            kwargs.update(
                {
                    "cells": self.cells,
                    "cell_data": self.cell_data,
                    "point_sets": self.point_sets,
                    "cell_sets": self.cell_sets,
                }
            )
        else:
            cell_data = {}
            if self.cell_data:
                for ic, c in enumerate(self.cells):
                    if c.type not in cell_data.keys():
                        cell_data[c.type] = {}
                    for k, v in self.cell_data.items():
                        cell_data[c.type][k] = v[ic]
            kwargs.update(
                {
                    "cells": self.cells_dict,
                    "cell_data": cell_data,
                    "node_sets": self.point_sets,
                }
            )

        return meshio.Mesh(**kwargs)

    def to_pyvista(self):
        """
        Convert mesh to :class:`pyvista.UnstructuredGrid`.

        Returns
        -------
        pyvista.UnstructuredGrid
            Output mesh.
        """
        try:
            import pyvista
            from ._common import (
                meshio_to_vtk_type,
                vtk_type_to_numnodes,
            )
        except ImportError:
            raise ModuleNotFoundError(
                "Converting to pyvista.UnstructuredGrid requires pyvista to be installed."
            )

        # Extract cells from toughio.Mesh object
        offset = []
        cells = []
        cell_type = []
        next_offset = 0
        for c in self.cells:
            vtk_type = meshio_to_vtk_type[c.type]
            numnodes = vtk_type_to_numnodes[vtk_type]
            offset += [next_offset + i * (numnodes + 1) for i in range(len(c.data))]
            cells.append(
                numpy.hstack((numpy.full((len(c.data), 1), numnodes), c.data)).ravel()
            )
            cell_type += [vtk_type] * len(c.data)
            next_offset = offset[-1] + numnodes + 1

        # Extract cell data from toughio.Mesh object
        cell_data = {k: numpy.concatenate(v) for k, v in self.cell_data.items()}

        # Create pyvista.UnstructuredGrid object
        points = self.points
        if points.shape[1] == 2:
            points = numpy.hstack((points, numpy.zeros((len(points), 1))))

        mesh = pyvista.UnstructuredGrid(
            numpy.array(offset),
            numpy.concatenate(cells),
            numpy.array(cell_type),
            numpy.array(points, numpy.float64),
        )

        # Set point data
        mesh.point_arrays.update(
            {k: numpy.array(v, numpy.float64) for k, v in self.point_data.items()}
        )

        # Set cell data
        mesh.cell_arrays.update(cell_data)

        return mesh

    def to_tough(self, filename="MESH", **kwargs):
        """
        Write TOUGH `MESH` file.

        Parameters
        ----------
        filename : str, optional, default 'MESH'
            Output file name.
        """
        self.write(filename, file_format="tough", **kwargs)

    def read_output(self, filename, time_step=-1):
        """
        Read TOUGH output file for a given time step.

        Parameters
        ----------
        filename : str
            Input file name.
        time_step : int, optional, default -1
            Data for given time step to import. Default is last time step.
        """
        from .. import read_output

        assert isinstance(time_step, int)

        out = read_output(filename)
        assert -len(out) <= time_step < len(out)

        _, labels, data = out[time_step]
        assert len(labels) == self.n_cells

        mapper = {k: v for v, k in enumerate(labels)}
        idx = [mapper[label] for label in numpy.concatenate(self.labels)]
        for k, v in data.items():
            self.cell_data[k] = self.split(v[idx])

    def write(self, filename, file_format=None, **kwargs):
        """
        Write mesh to file.

        Parameters
        ----------
        filename : str
            Output file name.
        file_format : str or None, optional, default None
            Output file format. If `None`, it will be guessed from file's
            extension. To write TOUGH MESH, `file_format` must be specified
            as 'tough' (no specific extension exists for TOUGH MESH).
        """
        from ._helpers import write

        write(filename, self, file_format, **kwargs)

    def plot(self, *args, **kwargs):
        """
        Display mesh using :method:`pyvista.UnstructuredGrid.plot``.
        """
        mesh = self.to_pyvista()
        mesh.plot(*args, **kwargs)

    def add_point_data(self, label, data):
        """
        Add a new point data array.

        Parameters
        ----------
        label : str
            Point data array name.
        data : array_like
            Point data array.
        """
        assert isinstance(label, str)
        assert isinstance(data, (list, tuple, numpy.ndarray))
        assert len(data) == self.n_points
        self.point_data[label] = numpy.asarray(data)

    def add_cell_data(self, label, data):
        """
        Add a new cell data array.

        Parameters
        ----------
        label : str
            Cell data array name.
        data : array_like
            Cell data array.
        """
        assert isinstance(label, str)
        assert isinstance(data, (list, tuple, numpy.ndarray))
        assert len(data) == self.n_cells
        self.cell_data[label] = self.split(data)

    def set_material(self, material, xlim=None, ylim=None, zlim=None):
        """
        Set material for cells within box selection defined by `xlim`,
        `ylim` and `zlim`.
        
        Parameters
        ----------
        material : str
            Material name.
        xlim : array_like or None, optional, default None
            Minimum and maximum values in X direction.
        ylim : array_like or None, optional, default None
            Minimum and maximum values in Y direction.
        zlim : array_like or None, optional, default None
            Minimum and maximum values in Z direction.

        Raises
        ------
        AssertionError
            If any input argument is not valid.
        """

        def isinbounds(x, bounds):
            return (
                numpy.logical_and(x >= min(bounds), x <= max(bounds))
                if bounds is not None
                else numpy.ones(len(x), dtype=bool)
            )

        assert isinstance(material, str)
        assert xlim is not None or ylim is not None or zlim is not None
        assert xlim is None or (
            isinstance(xlim, (list, tuple, numpy.ndarray)) and len(xlim) == 2
        )
        assert ylim is None or (
            isinstance(ylim, (list, tuple, numpy.ndarray)) and len(ylim) == 2
        )
        assert zlim is None or (
            isinstance(zlim, (list, tuple, numpy.ndarray)) and len(zlim) == 2
        )

        x, y, z = numpy.concatenate(self.centers).T
        mask_x = isinbounds(x, xlim)
        mask_y = isinbounds(y, ylim)
        mask_z = isinbounds(z, zlim)
        mask = numpy.logical_and(numpy.logical_and(mask_x, mask_y), mask_z)

        if mask.any():
            data = numpy.concatenate(self.cell_data["material"])
            imat = (
                self.field_data[material][0]
                if material in self.field_data.keys()
                else data.max() + 1
            )
            data[mask] = imat
            self.add_cell_data("material", data)
            self.field_data[material] = numpy.array([imat, 3])

    def near(self, point):
        """
        Return local index of cell nearest to query point.

        Parameters
        ----------
        point : array_like
            Coordinates of point to query.

        Returns
        -------
        tuple
            Local index of cell as a tuple (iblock, icell).
        """
        assert isinstance(point, (list, tuple, numpy.ndarray))
        assert numpy.ndim(point) == 1
        assert len(point) == self.points.shape[1]

        centers = numpy.concatenate(self.centers)
        idx = numpy.arange(self.n_cells)
        idx = idx[numpy.argmin(numpy.linalg.norm(centers - point, axis=1))]

        return get_local_index(self, idx)

    @property
    def points(self):
        """
        Coordinates of points.
        """
        return self._points

    @points.setter
    def points(self, value):
        self._points = value

    @property
    def cells(self):
        """
        Connectivity of cells.
        """
        if self._cells:
            return self._cells
        else:
            return [Cells(k, v) for k, v in self._cells_dict.items()]

    @cells.setter
    def cells(self, value):
        if isinstance(value, dict):
            self._cells = []
            self._cells_dict = value
        else:
            self._cells = [Cells(k, v) for k, v in value]
            self._cells_dict = {}

    @property
    def cells_dict(self):
        """
        Connectivity of cells (``meshio < 4.0.0``).
        """
        if self._cells:
            assert len(self._cells) == len(
                numpy.unique([c.type for c in self._cells])
            ), "More than one block of the same type. Cannot create dictionary."
            return dict(self._cells)
        else:
            return self._cells_dict

    @property
    def point_data(self):
        """
        Point data arrays.
        """
        return self._point_data

    @point_data.setter
    def point_data(self, value):
        self._point_data = value

    @property
    def cell_data(self):
        """
        Cell data arrays.
        """
        return self._cell_data

    @cell_data.setter
    def cell_data(self, value):
        self._cell_data = value

    @property
    def field_data(self):
        """
        Field data names.
        """
        return self._field_data

    @field_data.setter
    def field_data(self, value):
        self._field_data = value

    @property
    def point_sets(self):
        """
        Sets of points.
        """
        return self._point_sets

    @point_sets.setter
    def point_sets(self, value):
        self._point_sets = value

    @property
    def cell_sets(self):
        """
        Sets of cells.
        """
        return self._cell_sets

    @cell_sets.setter
    def cell_sets(self, value):
        self._cell_sets = value

    @property
    def n_points(self):
        """
        Number of points.
        """
        return len(self.points)

    @property
    def n_cells(self):
        """
        Number of cells.
        """
        return sum(len(c.data) for c in self.cells)

    @property
    def faces(self):
        """
        Connectivity of faces for each cell in mesh.
        """
        meshio_type_to_faces = {
            "tetra": {
                "triangle": numpy.array([[0, 1, 2], [0, 1, 3], [1, 2, 3], [0, 2, 3],]),
            },
            "pyramid": {
                "triangle": numpy.array([[0, 1, 4], [1, 2, 4], [2, 3, 4], [0, 3, 4],]),
                "quad": numpy.array([[0, 1, 2, 3],]),
            },
            "wedge": {
                "triangle": numpy.array([[0, 1, 2], [3, 4, 5],]),
                "quad": numpy.array([[0, 1, 3, 4], [1, 2, 4, 5], [0, 2, 3, 5],]),
            },
            "hexahedron": {
                "quad": numpy.array(
                    [
                        [0, 1, 2, 3],
                        [4, 5, 6, 7],
                        [0, 1, 4, 5],
                        [1, 2, 5, 6],
                        [2, 3, 6, 7],
                        [0, 3, 4, 7],
                    ]
                ),
            },
        }

        out = [
            [c[v] for v in meshio_type_to_faces[cell.type].values()]
            for cell in self.cells
            for c in cell.data
        ]

        # Convert to numpy.array
        arr = numpy.full((self.n_cells, 6, 4), -1)
        for i, x in enumerate(out):
            arr[i, : len(x[0]), : x[0].shape[1]] = x[0]
            if len(x) > 1:
                arr[i, len(x[0]) : len(x[0]) + len(x[1]), : x[1].shape[1]] = x[1]

        return self.split(arr)

    @property
    def labels(self):
        """
        Label of each cell in mesh.
        """
        from ._common import labeler

        return self.split([labeler(i) for i in range(self.n_cells)])

    @property
    def centers(self):
        """
        Center of each cell in mesh.
        """
        return [self.points[c.data].mean(axis=1) for c in self.cells]

    @property
    def connections(self):
        """
        Mesh connections assuming conformity and that points and cells are
        uniquely defined in mesh.

        Note
        ----
        Only for 3D meshes and first order cells.
        """
        assert (
            numpy.shape(self.points)[1] == 3
        ), "Connections for 2D mesh has not been implemented yet."

        # Reconstruct all the faces
        faces_dict = {"triangle": [], "quad": []}
        faces_cell = {"triangle": [], "quad": []}
        faces_index = {"triangle": [], "quad": []}
        numvert_to_face_type = {3: "triangle", 4: "quad"}

        for i, face in enumerate(numpy.concatenate(self.faces)):
            numvert = (face >= 0).sum(axis=-1)
            for j, (f, n) in enumerate(zip(face, numvert)):
                if n > 0:
                    face_type = numvert_to_face_type[n]
                    faces_dict[face_type].append(f[:n])
                    faces_cell[face_type].append(i)
                    faces_index[face_type].append(j)

        # Stack arrays or remove empty cells
        faces_dict = {
            k: numpy.sort(numpy.vstack(v), axis=1)
            for k, v in faces_dict.items()
            if len(v)
        }
        faces_cell = {k: v for k, v in faces_cell.items() if len(v)}
        faces_index = {k: v for k, v in faces_index.items() if len(v)}

        # Prune duplicate faces
        uf, tmp1, tmp2 = {}, {}, {}
        for k, v in faces_dict.items():
            up, uf[k] = numpy.unique(v, axis=0, return_inverse=True)
            tmp1[k] = [[] for _ in range(len(up))]
            tmp2[k] = [[] for _ in range(len(up))]

        # Make connections
        for k, v in uf.items():
            for i, j in enumerate(v):
                tmp1[k][j].append(faces_cell[k][i])
                tmp2[k][j].append(faces_index[k][i])
        conne = [vv for v in tmp1.values() for vv in v if len(vv) == 2]
        iface = [vv for v in tmp2.values() for vv in v if len(vv) == 2]

        # Reorganize output
        out = numpy.full((self.n_cells, 6), -1)
        for (i1, i2), (j1, j2) in zip(conne, iface):
            out[i1, j1] = i2
            out[i2, j2] = i1

        return self.split(out)

    @property
    def materials(self):
        """
        Material for each cell in mesh.
        """
        if "material" in self.cell_data.keys():
            if self.field_data:
                out = numpy.concatenate(self.cell_data["material"])
                try:
                    field_data_dict = {v[0]: k for k, v in self.field_data.items()}
                    return self.split([field_data_dict[mat] for mat in out])
                except KeyError:
                    logging.warning(
                        (
                            "field_data is not defined for all materials. "
                            "Returns materials as integers."
                        )
                    )
                    return self.cell_data["material"]
            else:
                return self.cell_data["material"]
        else:
            return self.split(numpy.ones(self.n_cells, dtype=int))

    @property
    def volumes(self):
        """
        Volumes for each cell in mesh.
        """

        def scalar_triple_product(a, b, c):
            c0 = b[:, 1] * c[:, 2] - b[:, 2] * c[:, 1]
            c1 = b[:, 2] * c[:, 0] - b[:, 0] * c[:, 2]
            c2 = b[:, 0] * c[:, 1] - b[:, 1] * c[:, 0]
            return a[:, 0] * c0 + a[:, 1] * c1 + a[:, 2] * c2

        meshio_type_to_tetra = {
            "tetra": numpy.array([[0, 1, 2, 3],]),
            "pyramid": numpy.array([[0, 1, 3, 4], [1, 2, 3, 4],]),
            "wedge": numpy.array([[0, 1, 2, 5], [0, 1, 4, 5], [0, 3, 4, 5],]),
            "hexahedron": numpy.array(
                [[0, 1, 3, 4], [1, 4, 5, 6], [1, 2, 3, 6], [3, 4, 6, 7], [1, 3, 4, 6],]
            ),
        }

        out = []
        for cell in self.cells:
            tetras = numpy.vstack(
                [c[meshio_type_to_tetra[cell.type]] for c in cell.data]
            )
            tetras = self.points[tetras]
            out.append(
                numpy.sum(
                    numpy.split(
                        numpy.abs(
                            scalar_triple_product(
                                tetras[:, 1] - tetras[:, 0],
                                tetras[:, 2] - tetras[:, 0],
                                tetras[:, 3] - tetras[:, 0],
                            )
                        ),
                        len(cell.data),
                    ),
                    axis=1,
                )
                / 6.0
            )
        return out


def from_meshio(mesh):
    """
    Convert a :class:`meshio.Mesh` to :class:`toughio.Mesh`.

    Parameters
    ----------
    mesh : meshio.Mesh
        Input mesh.

    Returns
    -------
    toughio.Mesh
        Output mesh.
    """
    if mesh.cell_data:
        version = get_meshio_version()
        if version[0] >= 4:
            cell_data = mesh.cell_data
        else:
            labels = numpy.unique(
                [kk for k, v in mesh.cell_data.items() for kk in v.keys()]
            )
            cell_data = {k: [] for k in labels}
            for k in cell_data.keys():
                for kk in mesh.cells.keys():
                    cell_data[k].append(mesh.cell_data[kk][k])

        for k in cell_data.keys():
            if k in meshio_data:
                cell_data["material"] = cell_data.pop(k)
                break
    else:
        cell_data = {}

    out = Mesh(
        points=mesh.points,
        cells=mesh.cells,
        point_data=mesh.point_data,
        cell_data=cell_data,
        field_data=mesh.field_data,
        point_sets=(
            mesh.point_sets
            if hasattr(mesh, "point_sets")
            else mesh.node_sets
            if hasattr(mesh, "node_sets")
            else None
        ),
        cell_sets=mesh.cell_sets if hasattr(mesh, "cell_sets") else None,
    )

    if "material" not in out.cell_data.keys():
        imat = (
            numpy.max([v[0] for v in mesh.field_data.values() if v[1] == 3]) + 1
            if mesh.field_data
            else 1
        )
        out.cell_data["material"] = out.split(numpy.full(out.n_cells, imat, dtype=int))
        out.field_data["dfalt"] = numpy.array([imat, 3])

    return out



import meshio

from . import avsucd, flac3d, pickle, tecplot, tough
from ._mesh import Mesh, from_meshio

__all__ = [
    "read",
    "write",
    "write_points_cells",
]


_extension_to_filetype = {
    ".dat": "tecplot",
    ".f3grid": "flac3d",
    ".pickle": "pickle",
}


def _filetype_from_filename(filename):
    """
    Determine file type from its extension.
    """
    import os

    ext = os.path.splitext(filename)[1].lower()
    return _extension_to_filetype[ext] if ext in _extension_to_filetype.keys() else ""


def read(filename, file_format=None, **kwargs):
    """
    Read unstructured mesh from file.

    Parameters
    ----------
    filename : str
        Input file name.
    file_format : str or None, optional, default None
        Input file format.

    Returns
    -------
    toughio.Mesh
        Imported mesh.
    """
    # Check file format
    assert isinstance(filename, str)
    fmt = file_format if file_format else _filetype_from_filename(filename)

    # Call custom readers
    format_to_reader = {
        "tough": (tough, (), {}),
        "avsucd": (avsucd, (), {}),
        "flac3d": (flac3d, (), {}),
        "pickle": (pickle, (), {}),
        "tecplot": (tecplot, (), {}),
    }
    if fmt in format_to_reader.keys():
        interface, args, default_kwargs = format_to_reader[fmt]
        _kwargs = default_kwargs.copy()
        _kwargs.update(kwargs)
        return interface.read(filename, *args, **_kwargs)
    else:
        mesh = meshio.read(filename, file_format)
        return from_meshio(mesh)


def write(filename, mesh, file_format=None, **kwargs):
    """
    Write unstructured mesh to file.

    Parameters
    ----------
    filename : str
        Output file name.
    mesh : toughio.Mesh
        Mesh to export.
    file_format : str or None, optional, default None
        Output file format.

    Other Parameters
    ----------------
    nodal_distance : str ('line' or 'orthogonal'), optional, default 'line'
        Only if ``file_format = "tough"``. Method to calculate connection
        nodal distances:
        - 'line': distance between node and common face along connecting
        line (distance is not normal),
        - 'orthogonal' : distance between node and its orthogonal
        projection onto common face (shortest distance).
    material_name : dict or None, default None
        Only if ``file_format = "tough"``. Rename cell material.
    material_end : str, array_like or None, default None
        Only if ``file_format = "tough"``. Move cells to bottom of block
        'ELEME' if their materials is in `material_end`.
    incon_eos : str or None, optional, default None
        Equation-of-state identifier to determine the actual number of
        primary variables to initialize. If `None`, TOUGH input `INCON`
        file will not be written.
    """
    # Check file format
    assert isinstance(filename, str)
    fmt = file_format if file_format else _filetype_from_filename(filename)

    # Call custom writer
    format_to_writer = {
        "tough": (
            tough,
            (),
            {
                "nodal_distance": "line",
                "material_name": None,
                "material_end": None,
                "incon_eos": None,
            },
        ),
        "avsucd": (avsucd, (), {}),
        "flac3d": (flac3d, (), {}),
        "pickle": (pickle, (), {}),
        "tecplot": (tecplot, (), {}),
    }
    if fmt in format_to_writer.keys():
        interface, args, default_kwargs = format_to_writer[fmt]
        _kwargs = default_kwargs.copy()
        _kwargs.update(kwargs)
        interface.write(filename, mesh, *args, **_kwargs)
    else:
        mesh = mesh.to_meshio()
        meshio.write(filename, mesh, file_format=file_format, **kwargs)


def write_points_cells(
    filename,
    points,
    cells,
    point_data=None,
    cell_data=None,
    field_data=None,
    file_format=None,
    **kwargs
):
    """
    Write unstructured mesh to file given points and cells data.

    Parameters
    ----------
    filename : str
        Output file name.
    points : ndarray
        Grid points array.
    cells : dict
        Grid cell data.
    point_data : dict or None, optional, default None
        Data associated to grid points.
    cell_data : dict or None, optional, default None
        Data associated to grid cells.
    field_data : dict or None, optional, default None
        Data names.
    file_format : str or None, optional, default None
        Output file format.

    Other Parameters
    ----------------
    kwargs : dict
        Refer to function ``write`` for additional information.
    """
    mesh = Mesh(
        points=points,
        cells=cells,
        point_data=point_data,
        cell_data=cell_data,
        field_data=field_data,
    )
    write(filename, mesh, file_format=file_format, **kwargs)



import meshio

from ._common import get_meshio_version
from ._helpers import read, write, write_points_cells
from ._mesh import Mesh, from_meshio

__all__ = [
    "read",
    "write",
    "write_points_cells",
    "Mesh",
    "from_meshio",
]



import os
import sys

from setuptools import find_packages, setup

base_dir = os.path.abspath(os.path.dirname(__file__))
about = {}
with open(os.path.join(base_dir, "toughio", "__about__.py"), "rb") as f:
    exec(f.read(), about)


DISTNAME = "toughio"
DESCRIPTION = "toughio"
LONG_DESCRIPTION = "toughio"
VERSION = about["__version__"]
AUTHOR = about["__author__"]
AUTHOR_EMAIL = about["__author_email__"]
URL = about["__website__"]
LICENSE = about["__license__"]
CLASSIFIERS = [
    "Programming Language :: Python",
    "Development Status :: 5 - Production/Stable",
    "License :: OSI Approved :: MIT License",
    "Natural Language :: English",
    "Operating System :: Microsoft :: Windows",
    "Operating System :: POSIX",
    "Operating System :: MacOS",
    "Intended Audience :: Science/Research",
    "Topic :: Scientific/Engineering",
]

# TODO: use new environment markers when dropping support for python < 3
# REQUIREMENTS = [
#     "meshio == 2.3.10; python_version <= '2.7'",
#     "meshio >= 3.0.0; python_version > '2.7'",
# ]
# EXTRA_REQUIREMENTS = {
#     "full": [
#         "pyvista == 0.22.4; python_version <= '2.7'",
#         "pyvista >= 0.23.1; python_version > '2.7'",
#     ],
# }
if sys.version_info < (3,):
    REQUIREMENTS = ["meshio == 2.3.10"]
    EXTRA_REQUIREMENTS = {"full": ["pyvista == 0.22.4"]}
else:
    REQUIREMENTS = ["meshio >= 3.0.0"]
    EXTRA_REQUIREMENTS = {"full": ["pyvista >= 0.23.1"]}
REQUIREMENTS += ["numpy >= 1.13.0"]
EXTRA_REQUIREMENTS["full"] += ["scipy>=0.9"]


if __name__ == "__main__":
    setup(
        name=DISTNAME,
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        author=AUTHOR,
        author_email=AUTHOR_EMAIL,
        url=URL,
        license=LICENSE,
        install_requires=REQUIREMENTS,
        extras_require=EXTRA_REQUIREMENTS,
        classifiers=CLASSIFIERS,
        version=about["__version__"],
        packages=find_packages(),
        include_package_data=True,
    )



from __future__ import division, unicode_literals, with_statement

import logging

import numpy

from .._common import meshio_data

__all__ = [
    "read",
    "write",
]


_eos_to_neq = {
    "eos1": 2,
    "eos2": 3,
    "eos3": 3,
    "eos4": 3,
    "eos5": 3,
    "eos7": 3,
    "eos8": 3,
    "eos9": 1,
    "eosmvoc": 4,
    "eoswasg": 4,
    "eco2n": 4,
    "eco2n_v2": 4,
    "eco2m": 4,
}


def block(keyword):
    """
    Decorator for block writing functions.
    """

    def decorator(func):
        from functools import wraps

        header = "----1----*----2----*----3----*----4----*----5----*----6----*----7----*----8"

        @wraps(func)
        def wrapper(f, *args):
            f.write("{}{}\n".format(keyword, header))
            func(f, *args)
            f.write("\n")

        return wrapper

    return decorator


def read(filename):
    """
    Read TOUGH MESH file is not supported yet. MESH file does not store
    any geometrical information except node centers.
    """
    raise NotImplementedError("Reading TOUGH MESH file is not supported.")


def write(filename, mesh, nodal_distance, material_name, material_end, incon_eos):
    """
    Write TOUGH MESH file.
    """
    assert nodal_distance in {"line", "orthogonal"}
    assert material_name is None or isinstance(material_name, dict)
    assert material_end is None or isinstance(material_end, (str, list, tuple))
    assert (
        incon_eos is None or incon_eos in _eos_to_neq.keys()
    ), "EOS '{}' is unknown or not supported.".format(incon_eos)

    # Required variables for blocks ELEME and CONNE
    num_cells = mesh.n_cells
    labels = numpy.concatenate(mesh.labels)
    nodes = numpy.concatenate(mesh.centers)
    materials = numpy.concatenate(mesh.materials)
    volumes = numpy.concatenate(mesh.volumes)
    boundary_conditions = (
        numpy.concatenate(mesh.cell_data["boundary_condition"])
        if "boundary_condition" in mesh.cell_data.keys()
        else numpy.zeros(num_cells, dtype=int)
    )
    points = mesh.points
    connections = numpy.concatenate(mesh.connections)
    gravity = numpy.array([0.0, 0.0, 1.0])

    # Define parameters related to faces
    faces = numpy.concatenate(mesh.faces)
    faces_dict, faces_cell = _get_faces(faces)
    face_normals, face_areas = _get_face_normals_areas(mesh, faces_dict, faces_cell)

    # Required variables for block INCON
    primary_variables = None
    porosities = None
    permeabilities = None
    if incon_eos:
        primary_variables = (
            numpy.concatenate(mesh.cell_data["initial_condition"])
            if "initial_condition" in mesh.cell_data.keys()
            else primary_variables
        )
        porosities = (
            numpy.concatenate(mesh.cell_data["porosity"])
            if "porosity" in mesh.cell_data.keys()
            else porosities
        )
        permeabilities = (
            numpy.concatenate(mesh.cell_data["permeability"])
            if "permeability" in mesh.cell_data.keys()
            else permeabilities
        )

    # Write MESH and INCON files
    write_buffer(
        filename,
        num_cells,
        labels,
        nodes,
        materials,
        volumes,
        boundary_conditions,
        points,
        connections,
        gravity,
        faces,
        face_normals,
        face_areas,
        primary_variables,
        porosities,
        permeabilities,
        nodal_distance,
        material_name,
        material_end,
        incon_eos,
    )


def write_buffer(
    filename,
    num_cells,
    labels,
    nodes,
    materials,
    volumes,
    boundary_conditions,
    points,
    connections,
    gravity,
    faces,
    face_normals,
    face_areas,
    primary_variables,
    porosities,
    permeabilities,
    nodal_distance,
    material_name,
    material_end,
    incon_eos,
):
    material_name = material_name if material_name else {}
    material_end = material_end if material_end else []
    material_end = [material_end] if isinstance(material_end, str) else material_end

    # Check INCON inputs and show warnings if necessary
    if incon_eos and primary_variables is None:
        logging.warning(
            ("Initial conditions are not defined. " "Writing INCON will be ignored.")
        )

    if porosities is not None:
        if not incon_eos:
            logging.warning("Porosity is only exported if incon_eos is provided.")
        else:
            assert (
                len(porosities) == num_cells and porosities.ndim == 1
            ), "Inconsistent porosity array."

    if permeabilities is not None:
        if not incon_eos:
            logging.warning(
                "Permeability modifiers are only exported if incon_eos is provided."
            )
        else:
            assert permeabilities.ndim in {1, 2}
            assert (
                permeabilities.shape == (num_cells, 3)
                if permeabilities.ndim == 2
                else len(permeabilities) == num_cells
            ), "Inconsistent permeability modifiers array."

    # Write MESH file
    with open(filename, "w") as f:
        _write_eleme(
            f,
            labels,
            nodes,
            materials,
            volumes,
            boundary_conditions,
            material_name,
            material_end,
        )
        _write_conne(
            f,
            labels,
            nodes,
            points,
            connections,
            gravity,
            boundary_conditions,
            faces,
            face_normals,
            face_areas,
            nodal_distance,
        )

    # Write INCON file
    if incon_eos and primary_variables is not None:
        import os

        head = os.path.split(filename)[0]
        with open(head + ("/" if head else "") + "INCON", "w") as f:
            _write_incon(
                f, incon_eos, labels, primary_variables, porosities, permeabilities,
            )


@block("ELEME")
def _write_eleme(
    f,
    labels,
    nodes,
    materials,
    volumes,
    boundary_conditions,
    material_name,
    material_end,
):
    """
    Write ELEME block.
    """
    # Apply time-independent Dirichlet boundary conditions
    volumes *= numpy.where(boundary_conditions, 1.0e50, 1.0)

    # Write ELEME block
    fmt = "{:5.5}{:>5}{:>5}{:>5}{:10.4e}{:>10}{:>10}{:10.3e}{:10.3e}{:10.3e}\n"
    ending = []
    iterables = zip(labels, materials, volumes, nodes)
    for label, material, volume, node in iterables:
        mat = material_name[material] if material in material_name.keys() else material
        record = fmt.format(
            label,  # ID
            "",  # NSEQ
            "",  # NADD
            mat,  # MAT
            volume,  # VOLX
            "",  # AHTX
            "",  # PMX
            node[0],  # X
            node[1],  # Y
            node[2],  # Z
        )
        if material not in material_end:
            f.write(record)
        else:
            ending.append(record)

    # Append ending cells at the end of the block
    if material_end:
        for record in ending:
            f.write(record)


@block("CONNE")
def _write_conne(
    f,
    labels,
    nodes,
    points,
    connections,
    gravity,
    boundary_conditions,
    faces,
    face_normals,
    face_areas,
    nodal_distance,
):
    """
    Write CONNE block.
    """
    # Define unique connection variables
    cell_list = set()
    clabels, centers, int_points, int_normals, areas, bounds = [], [], [], [], [], []
    for i, connection in enumerate(connections):
        if (connection >= 0).any():
            for iface, j in enumerate(connection):
                if j >= 0 and j not in cell_list:
                    # Label
                    clabels.append("{:5.5}{:5.5}".format(labels[i], labels[j]))

                    # Nodal points
                    centers.append([nodes[i], nodes[j]])

                    # Common interface defined by single point and normal vector
                    face = faces[i, iface]
                    int_points.append(points[face[face >= 0][0]])
                    int_normals.append(face_normals[i][iface])

                    # Area of common face
                    areas.append(face_areas[i][iface])

                    # Boundary conditions
                    bounds.append((boundary_conditions[i], boundary_conditions[j]))
        else:
            logging.warning(
                "Element '{}' is not connected to the grid.".format(labels[i])
            )
        cell_list.add(i)

    centers = numpy.array(centers)
    int_points = numpy.array(int_points)
    int_normals = numpy.array(int_normals)
    bounds = numpy.array(bounds)

    # Calculate remaining variables not available in itasca module
    lines = numpy.diff(centers, axis=1)[:, 0]
    isot = _isot(lines)
    angles = numpy.dot(lines, gravity) / numpy.linalg.norm(lines, axis=1)

    if nodal_distance == "line":
        fp = _intersection_line_plane(centers[:, 0], lines, int_points, int_normals)
        d1 = numpy.where(
            bounds[:, 0], 1.0e-9, numpy.linalg.norm(centers[:, 0] - fp, axis=1)
        )
        d2 = numpy.where(
            bounds[:, 1], 1.0e-9, numpy.linalg.norm(centers[:, 1] - fp, axis=1)
        )
    elif nodal_distance == "orthogonal":
        d1 = _distance_point_plane(centers[:, 0], int_points, int_normals, bounds[:, 0])
        d2 = _distance_point_plane(centers[:, 1], int_points, int_normals, bounds[:, 1])

    # Write CONNE block
    fmt = "{:10.10}{:>5}{:>5}{:>5}{:>5g}{:10.4e}{:10.4e}{:10.4e}{:10.3e}\n"
    iterables = zip(clabels, isot, d1, d2, areas, angles)
    for label, isot, d1, d2, area, angle in iterables:
        record = fmt.format(
            label,  # ID1-ID2
            "",  # NSEQ
            "",  # NAD1
            "",  # NAD2
            isot,  # ISOT
            d1,  # D1
            d2,  # D2
            area,  # AREAX
            angle,  # BETAX
        )
        f.write(record)


@block("INCON")
def _write_incon(
    f, incon_eos, labels, primary_variables, porosities, permeabilities,
):
    """
    Write INCON block.
    """
    # Check initial pore pressures
    cond = numpy.logical_and(
        primary_variables[:, 0] > -1.0e9, primary_variables[:, 0] < 0.0,
    )
    if cond.any():
        logging.warning("Negative pore pressures found in 'INCON'.")

    # Write label, porosity and permeability
    buffer = ["{:5.5}".format(label) for label in labels]

    if porosities is not None:
        buffer = [
            buf + "{:10}{:15.9e}".format("", phi)
            for buf, phi in zip(buffer, porosities)
        ]
    else:
        buffer = [buf + "{:10}{:15}".format("", "") for buf in buffer]

    if permeabilities is not None:
        if permeabilities.ndim == 1:
            buffer = [
                buf + "{:10.3e}{:10}{:10}".format(k, "", "")
                for buf, k in zip(buffer, permeabilities)
            ]
        else:
            buffer = [
                buf + "{:10.3e}{:10.3e}{:10.3e}".format(*k)
                for buf, k in zip(buffer, permeabilities)
            ]
    else:
        buffer = [buf + "{:10}{:10}{:10}".format("", "", "") for buf in buffer]

    # Write INCON block
    for pvar, buf in zip(primary_variables, buffer):
        if (pvar > -1.0e9).any():
            f.write(buf + "\n")
            for v in pvar:
                f.write("{:20.4e}".format(v) if v > -1.0e9 else "{:20}".format(""))
            f.write("\n")


def _get_faces(faces):
    """
    Return a face dictionary.
    """
    faces_dict = {"triangle": [], "quad": []}
    faces_cell = {"triangle": [], "quad": []}
    numvert_to_face_type = {3: "triangle", 4: "quad"}

    for i, face in enumerate(faces):
        numvert = (face >= 0).sum(axis=-1)
        for f, n in zip(face, numvert):
            if n > 0:
                face_type = numvert_to_face_type[n]
                faces_dict[face_type].append(f[:n])
                faces_cell[face_type].append(i)

    # Stack arrays or remove empty cells
    faces_dict = {
        k: numpy.sort(numpy.vstack(v), axis=1) for k, v in faces_dict.items() if len(v)
    }
    faces_cell = {k: v for k, v in faces_cell.items() if len(v)}

    return faces_dict, faces_cell


def _get_triangle_normals(mesh, faces, islice=[0, 1, 2]):
    """
    Calculate normal vectors of triangular faces.
    """

    def cross(a, b):
        return a[:, [1, 2, 0]] * b[:, [2, 0, 1]] - a[:, [2, 0, 1]] * b[:, [1, 2, 0]]

    triangles = numpy.vstack([c[islice] for c in faces])
    triangles = mesh.points[triangles]

    return cross(triangles[:, 1] - triangles[:, 0], triangles[:, 2] - triangles[:, 0],)


def _get_face_normals_areas(mesh, faces_dict, faces_cell):
    """
    Calculate face normal vectors and areas.
    """
    # Face normal vectors
    normals = numpy.concatenate(
        [_get_triangle_normals(mesh, v) for k, v in faces_dict.items()]
    )
    normals_mag = numpy.linalg.norm(normals, axis=-1)
    normals /= normals_mag[:, None]

    # Face areas
    areas = numpy.array(normals_mag)
    if len(faces_dict["quad"]):
        tmp = numpy.concatenate(
            [
                _get_triangle_normals(mesh, v, [0, 2, 3])
                if k == "quad"
                else numpy.zeros((len(v), 3))
                for k, v in faces_dict.items()
            ]
        )
        areas += numpy.linalg.norm(tmp, axis=-1)
    areas *= 0.5

    # Reorganize outputs
    face_normals = [[] for _ in range(mesh.n_cells)]
    face_areas = [[] for _ in range(mesh.n_cells)]
    iface = numpy.concatenate([v for v in faces_cell.values()])
    for i, normal, area in zip(iface, normals, areas):
        face_normals[i].append(normal)
        face_areas[i].append(area)

    return face_normals, face_areas


def _intersection_line_plane(center, lines, int_points, int_normals):
    """
    Calculate intersection point between a line defined by a point and a
    direction vector and a plane defined by one point and a normal vector.
    """
    tmp = _dot(int_points - center, int_normals) / _dot(lines, int_normals)
    return center + lines * tmp[:, None]


def _distance_point_plane(center, int_points, int_normals, mask):
    """
    Calculate orthogonal distance of a point to a plane defined by one
    point and a normal vector.
    """
    return numpy.where(mask, 1.0e-9, numpy.abs(_dot(center - int_points, int_normals)))


def _isot(lines):
    """
    Determine anisotropy direction given the direction of the line
    connecting the cell centers.
    Note
    ----
    It always returns 1 if the connection line is not colinear with X, Y or Z.
    """
    mask = lines != 0.0
    return numpy.where(mask.sum(axis=1) == 1, mask.argmax(axis=1) + 1, 1)


def _dot(A, B):
    """
    Custom dot product when arrays A and B have the same shape.
    """
    return (A * B).sum(axis=1)



from __future__ import division, with_statement

from copy import deepcopy

import numpy

from .._common import default
from ._helpers import block, check_parameters, dtypes

__all__ = [
    "read",
    "write",
]


def read(filename):
    """
    Read TOUGH input file.

    Parameters
    ----------
    filename : str
        Input file name.
    
    Returns
    -------
    dict
        TOUGH input parameters.
    """
    raise NotImplementedError("Reading TOUGH input file is not implemented yet.")


def write(filename, parameters):
    """
    Write TOUGH input file.

    Parameters
    ----------
    filename : str
        Output file name.
    parameters : dict
        Parameters to export.
    """
    from .._common import Parameters, default

    assert "rocks" in parameters.keys(), "Block 'ROCKS' (key 'rocks') is not defined."
    assert (
        "options" in parameters.keys()
    ), "Block 'PARAM' (key 'options') is not defined."

    params = deepcopy(Parameters)
    params.update(deepcopy(parameters))

    for k, v in default.items():
        if k not in params["default"].keys():
            params["default"][k] = v

    for rock in params["rocks"].keys():
        for k, v in params["default"].items():
            if k not in params["rocks"][rock].keys():
                params["rocks"][rock][k] = v

    buffer = write_buffer(params)
    with open(filename, "w") as f:
        for record in buffer:
            f.write(record)


@check_parameters(dtypes["PARAMETERS"])
def write_buffer(parameters):
    """
    Write TOUGH input file as a list of 80-character long record strings.
    """
    from .._common import eos, eos_select

    # Check that EOS is defined (for block MULTI)
    if parameters["isothermal"] and parameters["eos"] not in eos.keys():
        raise ValueError(
            "EOS '{}' is unknown or not supported.".format(parameters["eos"])
        )

    # Define input file contents
    out = ["{:80}\n".format(parameters["title"])]
    out += _write_rocks(parameters)
    out += _write_flac(parameters) if parameters["flac"] else []
    out += _write_multi(parameters) if parameters["eos"] else []
    out += _write_selec(parameters) if parameters["eos"] in eos_select else []
    out += _write_solvr(parameters) if parameters["solver"] else []
    out += _write_start() if parameters["start"] else []
    out += _write_param(parameters)
    out += _write_momop(parameters) if parameters["more_options"] else []
    out += _write_times(parameters) if parameters["times"] is not None else []
    out += _write_foft(parameters) if parameters["element_history"] is not None else []
    out += (
        _write_coft(parameters) if parameters["connection_history"] is not None else []
    )
    out += (
        _write_goft(parameters) if parameters["generator_history"] is not None else []
    )
    out += _write_gener(parameters) if parameters["generators"] else []
    out += _write_nover() if parameters["nover"] else []
    out += _write_endfi() if parameters["endfi"] else _write_endcy()
    return out


def _format_data(data):
    """
    Return a list of strings given input data and formats.
    """

    def to_str(x, fmt):
        x = "" if x is None or x == "" else x
        if isinstance(x, str):
            return fmt.replace("g", "").replace("e", "").format(x)
        else:
            return fmt.format(x)

    return [to_str(x, fmt) for x, fmt in data]


def _write_record(data):
    """
    Return a list with a single string.
    """
    return ["{:80}\n".format("".join(data))]


def _write_multi_record(data, ncol=8):
    """
    Return a list with multiple strings.
    """
    n = len(data)
    rec = [
        data[ncol * i : min(ncol * i + ncol, n)]
        for i in range(int(numpy.ceil(n / ncol)))
    ]
    return [_write_record(r)[0] for r in rec]


def _add_record(data, id_fmt="{:>5g}     "):
    """
    Return a list with a single string for additional records.
    """
    n = len(data["parameters"])
    rec = [(data["id"], id_fmt)]
    rec += [(v, "{:>10.3e}") for v in data["parameters"][: min(n, 7)]]
    return _write_record(_format_data(rec))


@check_parameters(dtypes["ROCKS"], keys="default")
@check_parameters(dtypes["MODEL"], keys=("default", "relative_permeability"))
@check_parameters(dtypes["MODEL"], keys=("default", "capillarity"))
@check_parameters(dtypes["ROCKS"], keys="rocks", is_list=True)
@check_parameters(
    dtypes["MODEL"], keys=("rocks", "relative_permeability"), is_list=True
)
@check_parameters(dtypes["MODEL"], keys=("rocks", "capillarity"), is_list=True)
@block("ROCKS", multi=True)
def _write_rocks(parameters):
    """
    TOUGH input ROCKS block data.
    
    Introduces material parameters for up to 27 different reservoir
    domains.
    """
    # Reorder rocks
    if parameters["rocks_order"] is not None:
        order = parameters["rocks_order"]
    else:
        order = parameters["rocks"].keys()

    out = []
    for k in order:
        # Load data
        data = default.copy()
        data.update(parameters["default"])
        data.update(parameters["rocks"][k])

        # Number of additional lines to write per rock
        nad = 0
        nad += 1 if data["relative_permeability"]["id"] is not None else 0
        nad += 1 if data["capillarity"]["id"] is not None else 0

        # Permeability
        per = data["permeability"]
        per = [per] * 3 if isinstance(per, float) else per
        assert isinstance(per, (list, tuple, numpy.ndarray)) and len(per) == 3

        # Record 1
        out += _write_record(
            _format_data(
                [
                    (k, "{:5.5}"),
                    (nad if nad else None, "{:>5g}"),
                    (data["density"], "{:>10.4e}"),
                    (data["porosity"], "{:>10.4e}"),
                    (per[0], "{:>10.4e}"),
                    (per[1], "{:>10.4e}"),
                    (per[2], "{:>10.4e}"),
                    (data["conductivity"], "{:>10.4e}"),
                    (data["specific_heat"], "{:>10.4e}"),
                ]
            )
        )

        # Record 2
        out += _write_record(
            _format_data(
                [
                    (data["compressibility"], "{:>10.4e}"),
                    (data["expansion"], "{:>10.4e}"),
                    (data["conductivity_dry"], "{:>10.4e}"),
                    (data["tortuosity"], "{:>10.4e}"),
                    (data["b_coeff"], "{:>10.4e}"),
                    (data["xkd3"], "{:>10.4e}"),
                    (data["xkd4"], "{:>10.4e}"),
                ]
            )
        )

        # Relative permeability
        out += _add_record(data["relative_permeability"]) if nad >= 1 else []

        # Capillary pressure
        out += _add_record(data["capillarity"]) if nad >= 2 else []
    return out


@check_parameters(dtypes["MODEL"], keys=("default", "permeability_model"))
@check_parameters(dtypes["MODEL"], keys=("default", "equivalent_pore_pressure"))
@check_parameters(dtypes["MODEL"], keys=("rocks", "permeability_model"), is_list=True)
@check_parameters(
    dtypes["MODEL"], keys=("rocks", "equivalent_pore_pressure"), is_list=True
)
@block("FLAC", multi=True)
def _write_flac(parameters):
    """
    TOUGH input FLAC block data (optional).

    Introduces mechanical parameters for each material in ROCKS block data.
    """
    # Reorder rocks
    if parameters["rocks_order"]:
        order = parameters["rocks_order"]
    else:
        order = parameters["rocks"].keys()

    # Record 1
    out = _write_record(
        _format_data(
            [
                (1 if parameters["creep"] else 0, "{:5g}"),
                (parameters["porosity_model"], "{:5g}"),
            ]
        )
    )

    # Additional records
    for k in order:
        # Load data
        data = default.copy()
        data.update(parameters["default"])
        data.update(parameters["rocks"][k])

        # Permeability law
        out += _add_record(data["permeability_model"], "{:>10g}")

        # Equivalent pore pressure
        out += _add_record(data["equivalent_pore_pressure"])
    return out


@block("MULTI")
def _write_multi(parameters):
    """
    TOUGH input MULTI block (optional).

    Permits the user to select the number and nature of balance equations
    that will be solved. The keyword MULTI is followed by a single data
    record. For most EOS modules, this data block is not needed, as
    default values are provided internally. Available parameter choices
    are different for different EOS modules.
    """
    from .._common import eos

    out = list(eos[parameters["eos"]])
    out[0] = parameters["n_component"] if parameters["n_component"] else out[0]
    out[1] = out[0] if parameters["isothermal"] else out[0] + 1
    out[2] = parameters["n_phase"] if parameters["n_phase"] else out[2]
    return [("{:>5d}" * len(out) + "\n").format(*out)]


@check_parameters(dtypes["SELEC"], keys="selections")
@block("SELEC")
def _write_selec(parameters):
    """
    TOUGH input SELEC block (optional).

    Introduces a number of integer and floating point parameters that are
    used for different purposes in different TOUGH modules (EOS7, EOS7R,
    EWASG, T2DM, ECO2N).
    """
    # Load data
    from .._common import select

    data = select.copy()
    data.update(parameters["selections"])

    # Record 1
    out = _write_record(_format_data([(v, "{:>5}") for v in data.values()]))

    # Record 2
    if parameters["extra_selections"] is not None:
        out += _write_multi_record(
            _format_data([(i, "{:>10.3e}") for i in parameters["extra_selections"]])
        )
    return out


@check_parameters(dtypes["SOLVR"], keys="solver")
@block("SOLVR")
def _write_solvr(parameters):
    """
    TOUGH input SOLVR block (optional).

    Introduces computation parameters, time stepping information, and
    default initial conditions.
    """
    from .._common import solver

    data = solver.copy()
    data.update(parameters["solver"])
    return _write_record(
        _format_data(
            [
                (data["method"], "{:1g}  "),
                (data["z_precond"], "{:>2g}   "),
                (data["o_precond"], "{:>2g}"),
                (data["rel_iter_max"], "{:>10.4e}"),
                (data["eps"], "{:>10.4e}"),
            ]
        )
    )


@block("START")
def _write_start():
    """
    TOUGH input START block (optional).

    A record with START typed in columns 1-5 allows a more flexible
    initialization. More specifically, when START is present, INCON data
    can be in arbitrary order, and need not be present for all grid blocks 
    (in which case defaults will be used). Without START, there must be a
    one-to-one correspondence between the data in blocks ELEME and INCON.
    """
    from .._common import header

    out = "{:5}{}\n".format("----*", header)
    return [out[:11] + "MOP: 123456789*123456789*1234" + out[40:]]


@check_parameters(dtypes["PARAM"], keys="options")
@check_parameters(dtypes["MOP"], keys="extra_options")
@block("PARAM")
def _write_param(parameters):
    """
    TOUGH input PARAM block data.
    
    Introduces computation parameters, time stepping information, and
    default initial conditions.
    """
    # Load data
    from .._common import options

    data = options.copy()
    data.update(parameters["options"])

    # Table
    if not isinstance(data["t_steps"], (list, tuple, numpy.ndarray)):
        data["t_steps"] = [data["t_steps"]]

    # Record 1
    from .._common import extra_options

    _mop = deepcopy(extra_options)
    _mop.update(parameters["extra_options"])
    mop = _format_data([(_mop[k], "{:>1g}") for k in sorted(_mop.keys())])
    out = _write_record(
        _format_data(
            [
                (data["n_iteration"], "{:>2g}"),
                (data["verbosity"], "{:>2g}"),
                (data["n_cycle"], "{:>4g}"),
                (data["n_second"], "{:>4g}"),
                (data["n_cycle_print"], "{:>4g}"),
                ("{}".format("".join(mop)), "{:>24}"),
                (None, "{:>10}"),
                (data["temperature_dependence_gas"], "{:>10.4e}"),
                (data["effective_strength_vapor"], "{:>10.4e}"),
            ]
        )
    )

    # Record 2
    out += _write_record(
        _format_data(
            [
                (data["t_ini"], "{:>10.4e}"),
                (data["t_max"], "{:>10.4e}"),
                (-((len(data["t_steps"]) - 1) // 8 + 1), "{:>9g}."),
                (data["t_step_max"], "{:>10.4e}"),
                (None, "{:>10g}"),
                (data["gravity"], "{:>10.4e}"),
                (data["t_reduce_factor"], "{:>10.4e}"),
                (data["mesh_scale_factor"], "{:>10.4e}"),
            ]
        )
    )

    # Record 2.1
    out += _write_multi_record(
        _format_data([(i, "{:>10.4e}") for i in data["t_steps"]])
    )

    # Record 3
    out += _write_record(
        _format_data(
            [
                (data["eps1"], "{:>10.4e}"),
                (data["eps2"], "{:>10.4e}"),
                (None, "{:>10.4e}"),
                (data["w_upstream"], "{:>10.4e}"),
                (data["w_newton"], "{:>10.4e}"),
                (data["derivative_factor"], "{:>10.4e}"),
            ]
        )
    )

    # Record 4
    n = len(data["incon"])
    out += _write_record(
        _format_data([(i, "{:>20.4e}") for i in data["incon"][: min(n, 4)]])
    )
    return out


@check_parameters(dtypes["MOMOP"], keys="more_options")
@block("MOMOP")
def _write_momop(parameters):
    """
    TOUGH input MOMOP block data (optional).

    Provides additional options.
    """
    from .._common import more_options

    _momop = more_options.copy()
    _momop.update(parameters["more_options"])
    out = _write_record(
        _format_data([(_momop[k], "{:>1g}") for k in sorted(_momop.keys())])
    )
    return out


@block("TIMES", multi=True)
def _write_times(parameters):
    """
    TOUGH input TIMES block data (optional).
    
    Permits the user to obtain printout at specified times.
    """
    data = parameters["times"]
    n = len(data)
    out = _write_record(_format_data([(n, "{:>5g}"),]))
    out += _write_multi_record(_format_data([(i, "{:>10.4e}") for i in data]))
    return out


@block("FOFT", multi=True)
def _write_foft(parameters):
    """
    TOUGH input FOFT block data (optional).
    
    Introduces a list of elements (grid blocks) for which time-dependent
    data are to be written out for plotting to a file called FOFT during
    the simulation.
    """
    return _write_multi_record(
        _format_data([(i, "{:>5g}") for i in parameters["element_history"]]), ncol=1
    )


@block("COFT", multi=True)
def _write_coft(parameters):
    """
    TOUGH input COFT block data (optional).
    
    Introduces a list of connections for which time-dependent data are to
    be written out for plotting to a file called COFT during the
    simulation.
    """
    return _write_multi_record(
        _format_data([(i, "{:>10g}") for i in parameters["connection_history"]]), ncol=1
    )


@block("GOFT", multi=True)
def _write_goft(parameters):
    """
    TOUGH input GOFT block data (optional).
    
    Introduces a list of sinks/sources for which time-dependent data are
    to be written out for plotting to a file called GOFT during the
    simulation.
    """
    return _write_multi_record(
        _format_data([(i, "{:>5g}") for i in parameters["generator_history"]]), ncol=1
    )


@check_parameters(dtypes["GENER"], keys="generators", is_list=True)
@block("GENER", multi=True)
def _write_gener(parameters):
    """
    TOUGH input GENER block data (optional).
    
    Introduces sinks and/or sources.
    """
    from .._common import generators

    out = []
    for k, v in parameters["generators"].items():
        # Load data
        data = generators.copy()
        data.update(v)

        # Table
        ltab, itab = None, None
        if data["times"] is not None and isinstance(
            data["times"], (list, tuple, numpy.ndarray)
        ):
            ltab, itab = len(data["times"]), 1
            assert isinstance(data["rates"], (list, tuple, numpy.ndarray))
            assert ltab > 1 and ltab == len(data["rates"])

        # Record 1
        out += _write_record(
            _format_data(
                [
                    (k, "{:5.5}"),
                    (None, "{:>5g}"),
                    (None, "{:>5g}"),
                    (None, "{:>5g}"),
                    (None, "{:>5g}"),
                    (ltab, "{:>5g}"),
                    (None, "{:>5g}"),
                    (data["type"], "{:4g}"),
                    (itab, "{:>1g}"),
                    (None if ltab else data["rates"], "{:>10.3e}"),
                    (None if ltab else data["specific_enthalpy"], "{:>10.3e}"),
                    (data["layer_thickness"], "{:>10.3e}"),
                ]
            )
        )

        # Record 2
        if ltab:
            out += _write_multi_record(
                _format_data([(i, "{:>14.7e}") for i in data["times"]]), ncol=4
            )

        # Record 3
        if ltab:
            out += _write_multi_record(
                _format_data([(i, "{:>14.7e}") for i in data["rates"]]), ncol=4
            )

        # Record 4
        if ltab and data["specific_enthalpy"] is not None:
            if isinstance(data["specific_enthalpy"], (list, tuple, numpy.ndarray)):
                specific_enthalpy = data["specific_enthalpy"]
            else:
                specific_enthalpy = numpy.full(ltab, data["specific_enthalpy"])
            out += _write_multi_record(
                _format_data([(i, "{:>14.7e}") for i in specific_enthalpy]), ncol=4
            )
    return out


@block("NOVER")
def _write_nover():
    """
    TOUGH input NOVER block data (optional).
    """
    return []


@block("ENDFI", noend=True)
def _write_endfi():
    """
    TOUGH input ENDFI block data (optional).
    """
    return []


@block("ENDCY", noend=True)
def _write_endcy():
    """
    TOUGH input ENDCY block data (optional).
    """
    return []



from string import ascii_uppercase

import meshio
import numpy

__all__ = [
    "vtk_to_meshio_type",
    "meshio_to_vtk_type",
    "vtk_type_to_numnodes",
    "meshio_type_to_ndim",
    "meshio_data",
    "get_meshio_version",
    "get_local_index",
    "labeler",
]


vtk_to_meshio_type = {
    0: "empty",
    1: "vertex",
    # 2: "poly_vertex",
    3: "line",
    # 4: "poly_line",
    5: "triangle",
    # 6: "triangle_strip",
    7: "polygon",
    # 8: "pixel",
    9: "quad",
    10: "tetra",
    # 11: "voxel",
    12: "hexahedron",
    13: "wedge",
    14: "pyramid",
    15: "penta_prism",
    16: "hexa_prism",
    21: "line3",
    22: "triangle6",
    23: "quad8",
    24: "tetra10",
    25: "hexahedron20",
    26: "wedge15",
    27: "pyramid13",
    28: "quad9",
    29: "hexahedron27",
    30: "quad6",
    31: "wedge12",
    32: "wedge18",
    33: "hexahedron24",
    34: "triangle7",
    35: "line4",
    #
    # 60: VTK_HIGHER_ORDER_EDGE,
    # 61: VTK_HIGHER_ORDER_TRIANGLE,
    # 62: VTK_HIGHER_ORDER_QUAD,
    # 63: VTK_HIGHER_ORDER_POLYGON,
    # 64: VTK_HIGHER_ORDER_TETRAHEDRON,
    # 65: VTK_HIGHER_ORDER_WEDGE,
    # 66: VTK_HIGHER_ORDER_PYRAMID,
    # 67: VTK_HIGHER_ORDER_HEXAHEDRON,
}
meshio_to_vtk_type = {v: k for k, v in vtk_to_meshio_type.items()}


vtk_type_to_numnodes = {
    0: 0,  # empty
    1: 1,  # vertex
    3: 2,  # line
    5: 3,  # triangle
    9: 4,  # quad
    10: 4,  # tetra
    12: 8,  # hexahedron
    13: 6,  # wedge
    14: 5,  # pyramid
    15: 10,  # penta_prism
    16: 12,  # hexa_prism
    21: 3,  # line3
    22: 6,  # triangle6
    23: 8,  # quad8
    24: 10,  # tetra10
    25: 20,  # hexahedron20
    26: 15,  # wedge15
    27: 13,  # pyramid13
    28: 9,  # quad9
    29: 27,  # hexahedron27
    30: 6,  # quad6
    31: 12,  # wedge12
    32: 18,  # wedge18
    33: 24,  # hexahedron24
    34: 7,  # triangle7
    35: 4,  # line4
}


meshio_type_to_ndim = {k: 3 for k in meshio_to_vtk_type.keys()}
meshio_type_to_ndim.update(
    {"empty": 0, "vertex": 1, "line": 2, "triangle": 2, "polygon": 2, "quad": 2,}
)


meshio_data = {
    "material",
    "avsucd:material",
    "flac3d:zone",
    "gmsh:physical",
    "medit:ref",
}


alpha = list(ascii_uppercase)
numer = ["{:0>2}".format(i) for i in range(100)]
nomen = ["{:1}".format(i + 1) for i in range(9)] + alpha


def get_meshio_version():
    """
    Return :module:`meshio` version as a tuple.

    Returns
    -------
    tuple
        :module:`meshio` version as tuple (major, minor, patch).
    """
    return tuple(int(i) for i in meshio.__version__.split("."))


def get_local_index(mesh, i):
    """
    Convert global cell index to local tuple index.

    Parameters
    ----------
    mesh : toughio.Mesh
        Input mesh.
    i : int
        Global cell index.
    """
    n_cells = numpy.cumsum([len(c.data) for c in mesh.cells])
    idx = numpy.nonzero(n_cells > i)[0][0]

    return (0, i) if not idx else (idx, i - n_cells[idx - 1])


def labeler(i):
    """
    Return five-character long cell labels following:
    - 1st: from A to Z
    - 2nd and 3rd: from 1 to 9 then A to Z
    - 4th and 5th: from 00 to 99
    Incrementation is performed from right to left.

    Parameters
    ----------
    i : int
        Cell general index in mesh.

    Returns
    -------
    str
        Label for cell `i`.

    Note
    ----
    Currently support up to 3,185,000 different cells.
    """
    q1, r1 = divmod(i, len(numer))
    q2, r2 = divmod(q1, len(nomen))
    q3, r3 = divmod(q2, len(nomen))
    _, r4 = divmod(q3, len(nomen))

    return "".join([alpha[r4], nomen[r3], nomen[r2], numer[r1]])
